/*

 suppt_c

 Shell Internal Support

 Date		Author		Comments
 ----		------		--------
 11/03/1998	A.D.Ives	New module added for v1.06
				Code to fetch a Shell Internal as a string
				brought out as a standalone function.
				New "$." Substitution - gets count of arguments.
				"%%" Now really gets a percent sign into
				the prompt.
				New "%[Name]" substitution in the prompt gets
				the value of an environment variable. (And
				"%[%Name] gets the value of a Shell Internal).
				New Shell Internals to return components of
				the current date and time, and details about
				the medium currently specified as the Data
				Default.

 1.07
 ----
 14/03/1998	A.D.Ives	New SI Internals to hold the FOR variables.
				New symbols "$*", "$^" to access them.
 15/03/1998			Includes sh_h to get PROG_VERSION.
				New prompt symbol %v gets version.
				Also, new IntVar %ver gets version.
				New String Operators: uc, lc, rev, part
				left, right, and mid.
				New escape character "^" at start of a
				command line will inhibit substitution
				of special symbols anywhere in the line.
 22/03/1998			New file functions to get the attributes
				of a named file.
				New math functions.

 1.08
 ----
 27/03/1998			New SI Internals to hold the DO variable.
				New symbol "$+" to access it.

*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <qdos.h>
#include <time.h>
#include "suppt.h"
#include "sh.h"
#include "libpjt.h"

extern int argCount ;		/* Count of command line arguments	*/
extern char **argStrings ;	/* Array of command line arguments	*/

/* Shell Internal Variables */

static long siRetCode ; 	/* Return code from last Job	*/
static long siResult ;		/* General purpose Result	*/
static char siLastKey , 	/* Last key hit on pause	*/
	    siForVar[128] ,	/* Holds value of FOR variable	*/
	    siForWc[128] ,	/* Holds partial FOR wildcard	*/
	    siDoVar[10] ;	/* Holds value of DO variable	*/

void ShellInit() {

  /* Initialise Shell Internals */

  siRetCode= siResult= siLastKey= 0 ;
  *siForVar= *siForWc= = *siDoVar= 0 ;

}

void *SI(int which, void *what) {

  /*

    Get/Set the value of a Shell Internal

    which	Code identifying which Internal we want.
    what	Pointer to the new value (or NULL if we're just getting
		a pointer to the value).

    Returns:	Void pointer to the relevant Internal.

  */

  void *res ;

  switch ( which ) {
  case SI_RETCODE :
    res= &siRetCode ;
    if ( what != NULL )
      siRetCode= (long) *((long *) (what)) ;
    break ;
  case SI_RESULT :
    res= &siResult ;
    if ( what != NULL )
      siResult= (long) *((long *) (what)) ;
    break ;
  case SI_LASTKEY :
    res= &siLastKey ;
    if ( what != NULL )
      siLastKey= (char) *((char *) (what)) ;
    break ;
  case SI_FOR :
    res= &siForVar ;
    if ( what != NULL )
      strcpy(siForVar, (char *) what) ;
    break ;
  case SI_FORW :
    res= &siForWc ;
    if ( what != NULL )
      strcpy(siForWc, (char *) what) ;
    break ;
  case SI_DO :
    res= &siDoVar ;
    if ( what != NULL )
      strcpy(siDoVar, (char *) what) ;
    break ;
  }
  return res ;
}

char *RPart(char *res, char *name, int which) {

  /*

    Return the nth (from the right-hand side)
    underscore-delimited part of a name.

  */

  int pos ;
  char buff[128], *ptr ;

  strcpy(buff, name) ;
  strrev(buff) ;
  ptr= buff ;
  if ( *buff == '_' )
    ptr++ ;
  Part(res, ptr, which) ;
  strrev(res) ;
  return res  ;
}

char *Part(char * res, char *name, int which) {

  /* Return the nth underscore-delimited part of a name */

  int pos ;
  char *ptr, buff[128] ;

  *res= 0 ;
  strcpy(buff, name) ;
  if ( *(buff + strlen(buff) - 1) != '_' )
    strcat(buff, "_") ;
  ptr= buff ;
  while ( *ptr ) {
    pos= strpos(ptr, '_') ;
    if ( pos == -1 )
      break ;
    if ( --which == 0 ) {
      strncpy(res, ptr, pos) ;
      *(res + pos)= 0 ;
      break ;
    }
    ptr+= (pos + 1) ;
  }
  return res ;
}

char *Month(char *name, char *res) {

  /* Convert a character month to two digits */

  char temp[4] ;

  strncpy(temp, name, 3) ;
  temp[3]= 0 ;
  strlwr(temp) ;
  if ( strcmp(temp, "jan") == 0 )
    strcpy(res, "01") ;
  else
  if ( strcmp(temp, "feb") == 0 )
    strcpy(res, "02") ;
  else
  if ( strcmp(temp, "mar") == 0 )
    strcpy(res, "03") ;
  else
  if ( strcmp(temp, "apr") == 0 )
    strcpy(res, "04") ;
  else
  if ( strcmp(temp, "may") == 0 )
    strcpy(res, "05") ;
  else
  if ( strcmp(temp, "jun") == 0 )
    strcpy(res, "06") ;
  else
  if ( strcmp(temp, "jul") == 0 )
    strcpy(res, "07") ;
  else
  if ( strcmp(temp, "aug") == 0 )
    strcpy(res, "08") ;
  else
  if ( strcmp(temp, "sep") == 0 )
    strcpy(res, "09") ;
  else
  if ( strcmp(temp, "oct") == 0 )
    strcpy(res, "10") ;
  else
  if ( strcmp(temp, "nov") == 0 )
    strcpy(res, "11") ;
  else
  if ( strcmp(temp, "dec") == 0 )
    strcpy(res, "12") ;
}

char *StringOp(char *name, char *res) {

  /* Do a string operation, and return the result */

  int dot, chars, ret, num, l, m ;
  char op[32], operand[128], buff[32], *ptr, *env ;
  qdirect_t this ;


  *res= 0 ;
  dot= strpos(name, ':') ;
  if ( dot > 0 && dot < strlen(name) - 1 ) {
    strncpy(op, name, dot) ;
    op[dot]= 0 ;
    m= strlen(op) ;
    ptr= name + dot + 1 ;		/* Get a pointer to the operand */
    if ( *ptr == '%' )

      /* Get the value of an internal name */

      Internal(operand, ptr) ;
    else
    if ( *ptr == '$' ) {

      /*

	Now let's really take a gamble, and try to get a
	standard substitution. (Recursion abounds!)

      */

      strcpy(operand, ptr ) ;
      SubstParams(operand, argCount, argStrings) ;
    }
    else {

      /* Might be an environment variable. Try it and see */

      env= getenv(ptr) ;
      if ( env != NULL )
	strcpy(operand, env) ;		/* Yep, we'll take it! */
      else
	strcpy(operand, ptr) ;		/* Nope, it's a literal (why?) */
    }
    l= strlen(operand) ;
    strlwr(op) ;

    /* File Attribute functions */

    if ( strcmp(op, "flen") == 0 ) {
      if ( (ret= qstat(operand, &this)) == 0 )
	sprintf(res, "%u", this.d_length) ;
      else
	sprintf(res, "E*%u", ret) ;
    }
    else
    if ( strcmp(op, "fupdt") == 0 ) {
      if ( (ret= qstat(operand, &this)) == 0 ) {
	cn_date(buff, this.d_update) ;
	sprintf(res, buff + 2) ;
      }
      else
	sprintf(res, "E*%u", ret) ;
    }
    else
    if ( strcmp(op, "fbkdt") == 0 ) {
      if ( (ret= qstat(operand, &this)) == 0 ) {
	cn_date(buff, this.d_backup) ;
	sprintf(res, buff + 2) ;
      }
      else
	sprintf(res, "E*%u", ret) ;
    }
    else
    if ( strcmp(op, "fvers") == 0 ) {
      if ( (ret= qstat(operand, &this)) == 0 )
	sprintf(res, "%u", this.u.v2.d_version) ;
      else
	sprintf(res, "E*%u", ret) ;
    }
    else
    if ( strcmp(op, "ftyp") == 0 ) {
      if ( (ret= qstat(operand, &this)) == 0 )
	sprintf(res, "%u", this.d_type) ;
      else
	sprintf(res, "E*%u", ret) ;
    }
    else
    if ( strcmp(op, "fdat") == 0 ) {
      if ( (ret= qstat(operand, &this)) == 0 )
	sprintf(res, "%u", this.d_datalen) ;
      else
	sprintf(res, "E*%u", ret) ;
    }
    else

    /* Maths ... */

    if ( strncmp(op, "add", 3) == 0 ) {
      num= atol(op + 4) ;
      sprintf(res, "%d", atol(operand) + num ) ;
    }
    else
    if ( strncmp(op, "sub", 3) == 0 ) {
      num= atol(op + 4) ;
      sprintf(res, "%d", atol(operand) - num ) ;
    }
    else
    if ( strncmp(op, "mul", 3) == 0 ) {
      num= atol(op + 4) ;
      sprintf(res, "%d", atol(operand) * num ) ;
    }
    else
    if ( strncmp(op, "div", 3) == 0 ) {
      num= atol(op + 4) ;
      sprintf(res, "%d", atol(operand) / num ) ;
    }
    else

    /* Strings ... */

    if ( strcmp(op, "len") == 0 )
      sprintf(res, "%u", strlen(operand)) ;
    else
    if ( strcmp(op, "uc") == 0 )
      strcpy(res, strupr(operand)) ;
    else
    if ( strcmp(op, "lc") == 0 )
      strcpy(res, strlwr(operand)) ;
    else
    if ( strcmp(op, "rev") == 0 )
      strcpy(res, strrev(operand)) ;
    else
    if ( strncmp(op, "part", 4) == 0 && m > 4 ) {
      char temp[128] ;
      chars= atoi(op + 4) ;
      if ( chars > 0 )
	Part(temp, operand, chars) ;
      else
      if ( chars < 0 )
	RPart(temp, operand, -chars) ;
      else
	*temp= 0 ;
      strcpy(res, temp) ;
    }
    else
    if ( strncmp(op, "left", 4) == 0 && m > 4 ) {
      chars= atoi(op + 4) ;
      if ( chars <= l ) {
	strncpy(res, operand, chars) ;
	res[chars]= 0 ;
      }
    }
    else
    if ( strncmp(op, "right", 5) == 0 && m > 5 ) {
      chars= atoi(op + 5) ;
      if ( chars <= l ) {
	strncpy(res, operand + l - chars, chars) ;
	res[chars]= 0 ;
      }
    }
    else
    if ( strncmp(op, "mid", 3) == 0 && m > 3 ) {

      int comma, start ;
      char op1[10], op2[10] ;

      comma= strpos(op, ',') ;
      if ( comma > 0 && comma < m ) {
	strncpy(op1, op + 3, comma - 3) ;
	op1[comma - 3]= 0 ;
	strcpy(op2, op + comma + 1) ;
	start= atoi(op1) ;
	chars= atoi(op2) ;
	if ( chars <= l ) {
	  strncpy(res, operand + start, chars) ;
	  res[chars]= 0 ;
	}
      }
    }
    else
    if ( strcmp(op, "daten") == 0 ) {

      /* Convert a date string into a simple serial number */

      if ( l == 24 ) {

	/* %longdate format */

	strncpy(res, operand + 20, 4 ) ;
	Month(operand + 4, res + 4) ;
	strncpy(res + 6, operand + 8, 2) ;
	strncpy(res + 8, operand + 11, 2) ;
	strncpy(res + 10, operand + 14, 2) ;
	strncpy(res + 12, operand + 17, 2) ;
	res[14]= 0 ;
      }
      else
      if ( l == 20 ) {
	strncpy(res, operand, 4) ;
	Month(operand + 5, res + 4) ;
	strncpy(res + 6, operand + 9, 2) ;
	strncpy(res + 8, operand + 12, 2) ;
	strncpy(res + 10, operand + 15, 2) ;
	strncpy(res + 12, operand + 18, 2) ;
	res[14]= 0 ;
      }
    }
  }
  return res ;
}

char *Internal(char *target, char *name) {

  /* Given the name of a Shell Internal, return it as a string */

  char Temp[128],
       sDate[32],
       sTime[32],
       Test[128],
       *what ;

  /* Do some setting up */

  *Temp= 0 ;
  what= name ;
  if ( *what == '%' )
    what++ ;		    /* No percent sign, thanks very much    */

  /* If there's a colon in there, it's a StringOp to be processed */

  if ( strpos(what, ':') != -1) {
    StringOp(what, Temp) ;
    strcpy(target, Temp) ;
    return target ;
  }

  strlwr(what) ;  /* Saves us calling stricmp()   */
  strcpy(Test, "_") ;
  strcat(Test, what) ;
  strcat(Test, "_") ; /* Built a special test string for the
			   medium-related stuff right at the
			   bottom of the if tree	  */

  /* Always get the latest date and time, in case we need it */

  sprintf(Temp, "%s", Date(sDate)) ;
  sprintf(Temp, "%s", Time(sTime)) ;

  if ( strcmp(what, "ver") == 0)
    strcpy(Temp, PROG_VERSION) ;
  else
  if ( strcmp(what, "data") == 0)
    getcwd(Temp, 64) ;
  else
  if ( strcmp(what, "prog") == 0)
    getcpd(Temp, 64) ;
  else
  if ( strcmp(what, "dest") == 0)
    getcdd(Temp, 64) ;
  else
  if ( strcmp(what, "retcode") == 0)
    sprintf(Temp, "%ld", siRetCode) ;
  else
  if ( strcmp(what, "result") == 0)
    sprintf(Temp, "%ld", siResult) ;
  else
  if ( strcmp(what, "lastkey") == 0)
    sprintf(Temp, "%u", siLastKey) ;
  else
  if ( strcmp(what, "longdate") == 0) {

    time_t t ;

    t= time(&t) ;
    strcpy(Temp,ctime(&t)) ;
    Temp[24]= 0 ;
  }
  else
  if ( strcmp(what, "date") == 0)
    sprintf(Temp, "%s", sDate) ;
  else
  if ( strcmp(what, "time") == 0)
    sprintf(Temp, "%s", sTime) ;
  else
  if ( strcmp(what, "weekday") == 0) {
    strncpy(Temp, sDate, 3) ;
    Temp[3]= 0 ;
  }
  else
  if ( strcmp(what, "month") == 0) {
    strncpy(Temp, sDate + 4, 3) ;
    Temp[3]= 0 ;
  }
  else
  if ( strcmp(what, "day") == 0) {
    strncpy(Temp, sDate + 8, 2) ;
    Temp[2]= 0 ;
  }
  else
  if ( strcmp(what, "year") == 0) {
    strncpy(Temp, sDate + 11, 4) ;
    Temp[4]= 0 ;
  }
  else
  if ( strcmp(what, "hour") == 0) {
    strncpy(Temp, sTime, 2) ;
    Temp[2]= 0 ;
  }
  else
  if ( strcmp(what, "minute") == 0) {
    strncpy(Temp, sTime + 3, 2) ;
    Temp[2]= 0 ;
  }
  else
  if ( strcmp(what, "second") == 0) {
    strncpy(Temp, sTime + 6, 2) ;
    Temp[2]= 0 ;
  }
  else
  if ( strstr("_name_alloc_free_total_kfree_ktotal_mfree_mtotal_", Test) != NULL ) {

    int res ;
    unsigned long alloc, free, total ;
    char mname[32] ;

    if ( (res= MediumInfo(&alloc, &free, &total, mname)) == 0 ) {
      if ( strcmp(what, "name") == 0 )
	strcpy(Temp, mname) ;
      else
      if ( strcmp(what, "alloc") == 0 )
	sprintf(Temp, "%u", alloc) ;
      else
      if ( strcmp(what, "free") == 0 )
	sprintf(Temp, "%u", free) ;
      else
      if ( strcmp(what, "total") == 0 )
	sprintf(Temp, "%u", total) ;
      else
      if ( strcmp(what, "kfree") == 0 )
	sprintf(Temp, "%u", (free * alloc) / 1024) ;
      else
      if ( strcmp(what, "ktotal") == 0 )
	sprintf(Temp, "%u", (total * alloc) / 1024) ;
      else
      if ( strcmp(what, "mfree") == 0 )
	sprintf(Temp, "%u", (free * alloc) / (1024 * 1024)) ;
      else
      if ( strcmp(what, "mtotal") == 0 )
	sprintf(Temp, "%u", (total * alloc) / (1024 * 1024)) ;
    }
    else
      sprintf(Temp, "E*%u", res) ;    /* Error! */
  }
  strcpy(target, Temp) ;
  return target ;
}

char *Prompt(char *PromptString, char *Text) {

  /*

    Process special characters in a prompt string, and copy the
    literal prompt into the supplied string.

    %d		Current Data Directory
    %p		Current Program Directory
    %D		Current Date
    %T		Current Time
    %[Name]	Environment Variable
    %[%Name]	Shell Internal

  */

  int Index ;
  char Dir[64], *Ptr, SwitchChar ;

  Ptr= PromptString ;
  Index= 0 ;
  while (*Ptr) {
    if (*Ptr == '%') {

      /* All symbols prefixed by % */

      SwitchChar= *++Ptr ;

      /* Which switch is it? */

      if ( SwitchChar == '[' ) {        /* %[Name] = Environment Variable */

	int EnvIndex ;
	char EnvName[128], Buff[128], *EnvPtr ;

	EnvIndex= 0 ;
	Ptr++ ;
	while ( ( *Ptr ) && ( *Ptr != ']' ) )
	  EnvName[EnvIndex++]= *Ptr++ ;
	EnvName[EnvIndex]= '\0' ;
	if ( EnvName[0] == '%' )	/* %[%Name] = Shell Internal	*/
	  EnvPtr= Internal(Buff, EnvName) ;
	else
	  EnvPtr= getenv(EnvName) ;
	if ( EnvPtr != NULL && *EnvPtr ) {
	  Text[Index]= '\0' ;
	  strcat(Text, EnvPtr) ;
	  Index+= strlen(EnvPtr) ;
      }
    }
    else
    if (SwitchChar == 'd') {          /* Data Directory */
      Text[Index]= '\0' ;
      getcwd(Dir, 64) ;
      strcat(Text, Dir) ;
      Index+= strlen(Dir) ;
    }
    else
    if (SwitchChar == 'p') {          /* Program Directory */
      Text[Index]= '\0' ;
      getcpd(Dir, 64) ;
      strcat(Text, Dir) ;
      Index+= strlen(Dir) ;
    }
    else
    if (SwitchChar == 'v') {          /* Shell version */
      Text[Index]= '\0' ;
      strcat(Text, PROG_VERSION) ;
      Index+= strlen(PROG_VERSION) ;
    }
    else
    if (SwitchChar == 'D') {          /* Date */
      Text[Index]= '\0' ;
      Date(Dir) ;
      strcat(Text, Dir) ;
      Index+= strlen(Dir) ;
    }
    else
    if (SwitchChar == 'T') {          /* Time */
      Text[Index]= '\0' ;
      Time(Dir) ;
      strcat(Text, Dir) ;
      Index+= strlen(Dir) ;
    }
    else
    if (SwitchChar == '%' )	    /* A percent sign! */
      Text[Index++]= '%' ;
      Ptr++ ;
    }
    else
      Text[Index++]= *Ptr++ ;
  }
  Text[Index]= '\0' ; /* Don't forget to terminate the string! */
  return Text ;
}

void SubstParams(char *Line, int argc, char *argv[]) {

  /*

    Substitute parameters in a string.
    Up to 26 parameters can be accessed, referenced as $A .. $Z.

    Line	    String to be processed
    argc	    Count of available parameters
    argv	    Array of parameter strings

  */

  int  Index,
       EnvIndex,
       ParNum ;
  char Temp[512],
       EnvName[128],
       Buff[128],
       *Ptr,
       *EnvPtr,
       SwitchChar ;

  if ( *Line == '^' ) {

    /* Special escape character prevents substitution */

    strcpy(Temp, Line + 1) ;
    strcpy(Line, Temp) ;
    return ;
  }

  Ptr= Line ;
  Index= 0 ;

  while ( *Ptr ) {
    if ( *Ptr == '$' ) {

      /* All parameters are prefixed by a $ sign */

      ParNum= 0 ;
      SwitchChar= *++Ptr ;

      if ( SwitchChar == '$' )

	/* $$ gets us a dollar sign !! */

	Temp[Index++]= '$' ;
      else
      if ( SwitchChar == '.' ) {

	/* $. gets us the total number of arguments */

	sprintf(Buff, "%u", argc - 1) ;
	Temp[Index]= 0 ;
	strcat(Temp, Buff) ;
	Index+= strlen(Buff) ;
      }
      else
      if ( SwitchChar == '*' ) {

	/* $* gets us the current FOR variable */

	Temp[Index]= 0 ;
	if ( *siForVar ) {
	  strcat(Temp, siForVar) ;
	  Index+= strlen(siForVar) ;
	}
	else {

	  /* No FOR Var set (yet) so pass it through */

	  strcat(Temp, "$*") ;
	  Index+= 2 ;
	}
      }
      else
      if ( SwitchChar == '^' ) {

	/* $^ gets us the current FOR wildcard */

	Temp[Index]= 0 ;
	if ( *siForWc ) {
	  strcat(Temp, siForWc) ;
	  Index+= strlen(siForWc) ;
	}
	else {

	  /* No FOR Var set (yet) so pass it through */

	  strcat(Temp, "$^") ;
	  Index+= 2 ;
	}
      }
      else
      if ( SwitchChar == '+' ) {

	/* $+ gets us the current DO variable */

	Temp[Index]= 0 ;
	if ( *siDoVar ) {
	  strcat(Temp, siDoVar) ;
	  Index+= strlen(siDoVar) ;
	}
	else {

	  /* No DO Var set (yet) so pass it through */

	  strcat(Temp, "$+") ;
	  Index+= 2 ;
	}
      }
      else
      if ( SwitchChar == '(' ) {

	/* $(NAME) gets us an environment variable) */

	EnvIndex= 0 ;
	Ptr++ ;
	while ( ( *Ptr ) && ( *Ptr != ')' ) ) {
	  EnvName[EnvIndex++]= *Ptr++ ;
	}
	EnvName[EnvIndex]= '\0' ;
	if ( EnvName[0] == '%' ) {

	  /* $(%NAME) gets us the value of a Shell Internal */

	  Temp[Index]= 0 ;
	  strcat(Temp, Internal(Buff, EnvName)) ;
	  Index+= strlen(Buff) ;
	}
	else {
	  EnvPtr= getenv(EnvName) ;
	  if ( EnvPtr != NULL ) {
	    Temp[Index]= '\0' ;
	    strcat(Temp, EnvPtr) ;
	    Index+= strlen(EnvPtr) ;
	  }
	}
      }
      else {

	/* Otherwise, identify which parameter we want */

	if ((SwitchChar >= 'A') && (SwitchChar <= 'Z'))
	  ParNum= (SwitchChar - 'A') + 1 ;
	if ((SwitchChar >= 'a') && (SwitchChar <= 'z'))
	  ParNum= (SwitchChar - 'a') + 1 ;

	if ((ParNum > 0) && (ParNum <= argc)) {
	  Temp[Index]= '\0' ;
	  strcat(Temp, argv[ParNum]) ;
	  Index+= strlen(argv[ParNum]) ;
	}
      }
      if ( *Ptr )
	Ptr++ ;
    } else
    Temp[Index++]= *Ptr++ ;
  }
  Temp[Index]= '\0' ; /* Don't forget to terminate the string! */
  strcpy(Line, Temp) ;
}
